;;; mixsim.el --- MIX machine simulator

;; Copyright (C) 2000 Free Software Foundation, Inc.

;; Authors: 2000 Sergey Poznyakoff
;; Version:  1.0
;; Keywords: mix TAOCP languages
;; $Id: mixsim.el,v 1.2 2005/04/09 18:16:50 gray Exp $

;; This file is part of MIX simulator.

;; MIX simulator is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; MIX simulator is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with MIX simulator; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;; Memory
(defconst +MEMSIZE+ 4000)
(defconst +NUMUNITS+ 20)
(defconst +BYTESIZE+ 6)
(defconst +BYTESPERWORD+ 5)
(defconst +WORDSIZE+ (* +BYTESIZE+ +BYTESPERWORD+))
(defconst +BASE+ 64)
(defconst +JA+ 40)
(defvar *mix-cell* (make-vector +MEMSIZE+  0))
;; Registers
(defvar *mix-reg* (make-vector 10 0))
(defvar *mix-compi* 0)      ;; Comparison indicator
(defvar *mix-ovtog* nil)    ;; Overflow toggle

(defconst +mix-alpha+
  " ABCDEFGHInJKLMNOPQRepSTUVWXYZ0123456789.,()+-*/=$<>@;:'abcdfghi")


(defconst :op-fun 0)
(defconst :op-time 1)
(defconst :op-iflag 2)

(defconst +mix-optable+
  '[[mixop-nop     1 nil]
    [mixop-add     2 nil]
    [mixop-sub     2 nil]
    [mixop-mul    10 nil]
    [mixop-div    12 nil]
    [mixop-spec    1 nil]
    [mixop-shift   2 nil]
    [mixop-move    1 nil]
    [mixop-load    2 nil]
    [mixop-load    2 t]
    [mixop-load    2 t]
    [mixop-load    2 t]
    [mixop-load    2 t]
    [mixop-load    2 t]
    [mixop-load    2 t]
    [mixop-load    2 nil]
    [mixop-loadn   2 t]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 nil]
    [mixop-loadn   2 t]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-store   2 nil]
    [mixop-jbus    1 nil]
    [mixop-ioc     1 nil]
    [mixop-in      1 nil]
    [mixop-out     1 nil]
    [mixop-jred    1 nil]
    [mixop-jump    1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-regjump 1 nil]
    [mixop-addrop  1 nil]
    [mixop-addrop  1 t]
    [mixop-addrop  1 t]
    [mixop-addrop  1 t]
    [mixop-addrop  1 t]
    [mixop-addrop  1 t]
    [mixop-addrop  1 t]
    [mixop-addrop  1 nil]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]
    [mixop-compare 2 t]])

;; I/O units
(defconst +TAPE0+ 0)
(defconst +TAPE7+ 7)
(defconst +DISK8+ 8)
(defconst +DISK15+ 15)
(defconst +READER+ 16)
(defconst +PUNCH+  17)
(defconst +PRINTER+ 18)
(defconst +TTY+ 19)
(defconst :io-ufn 0)
(defconst :io-time 1)
(defconst :io-position 2)
(defconst :io-seek-pos 3)
(defconst :io-tape-end 4)
(defconst :io-seek-time 5)
(defconst :io-address 6)
(defconst :io-op 7)
(defconst :io-clock 8)
(defconst :io-complete 9)
(defconst :io-path 10)
(defconst :io-bufname 11)
(defconst :io-buffer 12)

(defconst +IOC+ 0)
(defconst +IN+ 1)
(defconst +OUT+ 2)

(defvar *mix-unit*
  ;; DevFn     iot  pos skp end skt addr op clock compl  path    bufname buffer
  '[[mix-tape 50    0  0  0    50  0   0   -1    t    "tape0"  "tape-0"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape1"  "tape-1"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape2"  "tape-2"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape3"  "tape-3"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape4"  "tape-4"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape5"  "tape-5"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape6"  "tape-6"  nil]
    [mix-tape 50    0  0  0    50  0   0   -1    t    "tape7"  "tape-7"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk0"  "disk-0"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk1"  "disk-1"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk2"  "disk-2"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk3"  "disk-3"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk4"  "disk-4"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk5"  "disk-5"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk6"  "disk-6"  nil]
    [mix-disk 250   0  0  0     1  0   0   -1    t    "disk7"  "disk-7"  nil]
    [mix-reader 10000 0  0  0   0  0   0   -1    t    "deck"   "reader"  nil]
    [mix-punch  20000 0  0  0   0  0   0   -1    t    "stderr" "puncher" nil]
    [mix-printer 7500 0  0  0 10000  0 0   -1    t    "stdout" "printer" nil]
    [mix-tty     0    0  0  0     0  0 0   -1    t    "stdin"  "tty"     nil]])
;; Current I/O unit
(defvar *mix-current-unit*)

(defun mix-select-unit (unum)
  (setq *mix-current-unit* (aref *mix-unit* unum)))

(defun mix-get-unit-parm (num)
  (aref *mix-current-unit* num))

(defun mix-set-unit-parm (num value)
  (aset *mix-current-unit* num value))

(defun mix-select-unit (n)
  (setq *mix-current-unit* (aref *mix-unit* n)))

(defun mix-get-unit-buffer ()
  (or
   (mix-get-unit-parm :io-buffer)
   (let* ((bufname (concat "*mix-"
			   (mix-get-unit-parm :io-bufname)
			   "*"))
	  (buffer (get-buffer-create bufname))
	  (path (mix-get-unit-parm :io-path)))
     (save-excursion
       (set-buffer buffer)
       (erase-buffer)
       (cond
	((and (file-exists-p path) (file-readable-p path))
	 (insert-file-contents path))
	(t
	 (set-visited-file-name path))))
     (mix-set-unit-parm :io-buffer buffer))))
     
     
(defvar *mix-loc*)  ;; Location of the next instruction
(defvar *mix-inst*) ;; Current instruction
(defvar *mix-op*)   ;; Opcode of the current instruction
(defvar *mix-mem*)  ;; Address of the present instruction plus indexing
(defvar *mix-fld*)  ;; Field spec of the current instruction

;; work registers
(defvar *mix-rA*)
(defvar *mix-rX*)
(defvar *mix-rI1*)
(defvar *mix-rI2*)

(defvar *mix-uclock*)  ;; Simulated execution time
(defvar *mix-instime*) ;; Time of previous instruction

;; General-purpose macros
(defmacro inc (var)
  (list 'setq var (list '1+ var)))

(defmacro dec (var)
  (list 'setq var (list '1- var)))

(defmacro mix-fill-array (array size val)
  (` (let ((i 0))
       (while (< i (, size))
	 (aset (, array) i (, val))
	 (inc i)))))

;; Register access macros
;ADD I F C
;1 2 3 4 5
(defmacro mix-address (word)
  (` (+ (* (mix-byte (, word) 1) +BASE+)
	(mix-byte (, word) 0))))

(defmacro mix-index (word)
  (` (mix-byte (, word) 3)))

(defmacro mix-field (word)
  (` (mix-byte (, word) 4)))

(defmacro mix-code (word)
  (` (mix-byte (, word) 5)))

(defconst :rA  0)
(defconst :rI1 1)
(defconst :rI2 2)
(defconst :rI3 3)
(defconst :rI4 4)
(defconst :rI5 5)
(defconst :rI6 6)
(defconst :rX  7)
(defconst :rJ  8)
(defconst :rZERO 9)

(defmacro mix-get-reg (n)
  (` (aref *mix-reg* (, n))))

(defmacro mix-set-reg (n val)
  (` (aset *mix-reg* (, n) (, val))))

(defun mix-set-reg-magn (n val)
  (let ((i 1))
    (while (<= i 5)
      (aset (aref *mix-reg* n) i (aref val i))
      (inc i))))

(defmacro mix-zero ()
  (` (aref *mix-reg* :rZERO)))


;; Translate character c to MIX representation
(defun mix-xlat-to-mix (c)
  (let ((pos 0)
	(len (length +mix-alpha+)))
    (while (and
	    (< pos len)
	    (not (eq c (aref +mix-alpha+ pos))))
      (inc pos))
    pos))

(defun mix-xlat-to-ascii (n)
  (cond
   ((< n (length +mix-alpha+))
    (aref +mix-alpha+ n))
   (t
    ??)))

;; Cell fiddling
(defmacro mix-sign (word)
  (` (aref (, word) 0)))

;; Get the address field out of a word.
;; Returns signed integer
(defun mix-get-addr (value)
  (+
   (* (mix-byte value 1) +BASE+)
   (mix-byte value 2)))

(defun mix-get-fld (value)
  (mix-byte value 4))

(defun mix-get-code (value)
  (mix-byte value 5))

(defmacro set-sign (word sign)
  (` (aset (, word) 0 (, sign))))

(defmacro mix-byte (word n)
  (` (aref (, word) (, n))))

(defmacro set-byte (word n value)
  (` (aset (, word) (, n) (, value)))) 

(defmacro mix-cell (loc)
  (` (aref *mix-cell* (, loc))))

(defun mix-set-cell (address word)
  (aset *mix-cell* address word))

(defun mix-get-cell (address)
  (aref *mix-cell* address))

;;; 64-cimal arithmetics
;;;

(defun mix-copy-sign (a b)
  (set-sign a (mix-sign b)))

;; Create a MIX-number LEN bytes long
(defun mix-make-number (len)
  (let ((res (make-vector (1+ len) 0)))
    (set-sign res 1)
    res))

;; Converts the the MIX-word WORD to integer number.
;; Optional parameter OPT-LEN specifies the length of the word.
;; Note possible overflow
(defun mix-convert-word-to-number (word &optional opt-len)
  (let* ((len (or opt-len 5))
	 (i 1)
	 (value 0))
    (while (<= i len)
      (setq value (+ (* value +BASE+) (mix-byte word i)))
      (inc i))
    (if (= (mix-sign word) -1)
	(- value)
      value)))

;; Converts the integer number VALUE to the MIX-word.
;; Optional parameter OPT-LEN specifies the length of the resulting word
(defun mix-convert-to-word (value &optional opt-len)
  (let* ((len (or opt-len 5))
	 (num (mix-make-number len))
	 val)
    (cond
     ((< value 0)
      (setq val (- value))
      (set-sign num -1))
     (t
      (setq val value)))
    (let ((i len))
      (while (> i 0)
	(let ((q (% val +BASE+)))
	  (set-byte num i q)
	  (setq val (/ val +BASE+)))
	(dec i)))
    num))

;; Copy the MIX-number SRC which is SRC-LEN bytes long into
;; DST DST-LEN bytes long, truncating it if necessary.
;; Preserve the sign.
(defun mix-setq (dst dst-len src src-len)
  (mix-copy-sign dst src)
  (let ((offs (- src-len dst-len))
	(i dst-len))
    (while (> i 0)
      (set-byte dst i (cond
		       ((> (+ offs i) 0)
			(mix-byte src (+ offs i)))
		       (t
			0)))
      (dec i)))
  dst)

;; Copy the MIX-number A which is LENGTH bytes long into
;; a MIX-number COUNT bytes long, truncating A if necessary.
;; Preserve the sign.
(defun mix-copy-number (a length count)
  (let ((res (mix-make-number count)))
    (mix-copy-sign res a)
    (let ((offs (- length count))
	  (i count))
      (while (and (> i 0)
		  (> (+ offs i) 0))
	(set-byte res i (mix-byte a (+ offs i)))
	(dec i)))
    res))

;; Return the absolute value of the word
(defun mix-abs (word)
  (let ((r (mix-copy-number word 5 5)))
    (set-sign r 1)
    r))

;; Basic comparison predicates
(defun mix-zero-p (word &optional optlen)
  (let ((i (or optlen 5)))
    (while (and (> i 0) (= (mix-byte word i) 0))
      (dec i))
    (= i 0)))

(defun mix-positive-p (word)
  (and
   (not (mix-zero-p word))
   (= (mix-sign word) 1)))

(defun mix-negative-p (word)
  (and
   (not (mix-zero-p word))
   (= (mix-sign word) -1)))

(defun mix-nonnegative-p (word)
  (not (mix-negative-p word)))

(defun mix-nonpositive-p (word)
  (not (mix-positive-p word)))

;;; Unsigned arithmetics
;; Comparison
(defun mix-unsigned-cmp (op a b &optional optlen)
  (let ((len (or optlen 5))
	(n 1))
    (while (and (< n len)
		(= (mix-byte a n)
		   (mix-byte b n)))
      (inc n))
    (funcall op
	     (mix-byte a n)
	     (mix-byte b n))))

(defun mix-unsigned-eq (a b &optional optlen)
  (mix-unsigned-cmp '= a b optlen))

(defun mix-unsigned-ne (a b &optional optlen)
  (not (mix-unsigned-eq a b optlen)))

(defun mix-unsigned-gt (a b &optional optlen)
  (mix-unsigned-cmp '> a b optlen))

(defun mix-unsigned-ge (a b &optional optlen)
  (mix-unsigned-cmp '>= a b optlen))

(defun mix-unsigned-lt (a b &optional optlen)
  (mix-unsigned-cmp '< a b optlen))

(defun mix-unsigned-le (a b &optional optlen)
  (mix-unsigned-cmp '<= a b optlen))

;; Addition / Subtraction
(defun mix-unsigned-add (a b &optional optlen)
  (let* ((len (or optlen 5))
	 (n len)
	 (c (mix-make-number len)))
    (while (> n 0)
      (let* ((sum (+ (mix-byte c n)
		     (mix-byte a n)
		     (mix-byte b n))))
	(if (> sum (1- +BASE+))
	    (progn
	      (setq sum (- sum +BASE+))
	      (if (> n 1)
		  (set-byte c (1- n) 1)
		(setq *mix-ovtog* +ON+))))
	(set-byte c n sum))
      (dec n))
    c))

;; NOTE: a > b
(defun mix-unsigned-sub (a b &optional optlen)
  (let* ((len (or optlen 5))
	 (n len)
	 (atmp (mix-copy-number a len len))
	 (c (mix-make-number len)))
    (while (> n 0)
      (let ((dif (- (mix-byte atmp n)
		    (mix-byte b n))))
	(cond
	 ((< dif 0)
	  (setq dif (+ +BASE+ dif))
	  (let ((i (1- n)))
	    (while (and (> i 0)
			(< (set-byte atmp i (1- (mix-byte atmp i))) 0))
	      (set-byte atmp i (+ (mix-byte atmp i) +BASE+))
	      (dec i)))))
	(set-byte c n dif))
      (dec n))
    c))

;; Multiplication / Division

;; Shift left a MIX-nuber A
(defun mix-lsh (a count &optional optlen)
  (let* ((len (or optlen 5))
	 (c (mix-make-number len))
	 (i len))
    (if (< count len)
	(while (> i count)
	  (set-byte c (- i count) (mix-byte a i))
	  (dec i)))
    (set-sign c (mix-sign a))
    c))

(defun mix-lcsh (a count &optional optlen)
  (let* ((len (or optlen 5))
	 (doublelen (* 2 len))
	 (c (mix-lsh (mix-copy-number a len doublelen)
		     (% count len) doublelen))
	 (temp (mix-unsigned-add
		(mix-copy-number (mix-rsh c len doublelen) doublelen len)
		(mix-copy-number c doublelen len)
		len)))
    (mix-copy-sign temp a)
    temp))
		
	 

(defun mix-rsh (a count &optional optlen)
  (let* ((len (or optlen 5))
	 (c (mix-make-number len))
	 (i 1))
    (if (< count len)
	(while (<= (+ i count) len)
	  (set-byte c (+ i count) (mix-byte a i))
	  (inc i)))
    (set-sign c (mix-sign a))
    c))

;; Multiply a MIX-number A by DIGIT.
(defun mix-one-mul (a digit &optional optlen)
  (let* ((len (or optlen 5))
	 (res (mix-make-number len))
	 (carry 0))
    (while (> len 0)
      (let ((value (+ (* (mix-byte a len) digit) carry)))
	(set-byte res len (% value +BASE+))
	(setq carry (/ value +BASE+)))
      (dec len))
    (if (> carry 0)
	(setq *mix-ovtog* +ON+))
    (set-sign res (mix-sign a))
    res))


;; Unsigned multiplication of two MIX-numbers.
;; Result is always positive.
(defun mix-unsigned-mul (an bn &optional opt-prec)
  (let* ((prec (or opt-prec 10))
	 (a (mix-copy-number an 5 prec))
	 (b (mix-copy-number bn 5 prec))
	 (c (mix-make-number prec))
	 (i prec))
    (while (> i 0)
      (setq c (mix-unsigned-add c
				(mix-lsh
				 (mix-one-mul a (mix-byte b i) prec)
				 (- prec i)
				 prec)
				prec))
      (dec i))
    c))

;; Position of first nonzero byte in a word
(defun mix-number-start (num len)
  (let ((n 1))   
    (while (and (< n len)
		(= (mix-byte num n) 0))
      (inc n))
    n))

;; Count the actual length of NUM
(defun mix-count-bytes (num &optional length)
  (let ((len (or length 5))
	(pos 1))
    (while (and (< pos len)
		(= (mix-byte num pos) 0))
      (inc pos))
    (1+ (- len pos))))
    
(defun test-guess (num1 num2 qdig qguess n2ptr)
  (cond
   ((= n2ptr 10)
    nil)
   (t
    (> (* (mix-byte num2 (1+ n2ptr)) qguess)
       (+ (* +BASE+
	     (-
	      (+ (* (mix-byte num1 qdig) +BASE+)
		 (mix-byte num1 (1+ qdig)))
	      (* (mix-byte num2 n2ptr) qguess)))
	  (mix-byte num1 (+ qdig 2)))))))

;; Divide the long word (10 bytes) A by short word B
;; Return 5-byte quotient.
;; 
(defun mix-unsigned-div (a b)
  (cond
   ((or (mix-zero-p b)
	(mix-unsigned-ge (mix-copy-number (mix-rsh a 5 10) 10 5) b))
    (setq *mix-ovtog* 1)
    (mix-copy-number a 10 5))
   ((or (mix-zero-p a 10)
	(mix-unsigned-eq b '[1 0 0 0 0 1])
	(mix-unsigned-eq b '[-1 0 0 0 0 1]))
    (mix-copy-number a 10 5))
   (t
    ;; Set up the divide
    (let* ((num1 (mix-copy-number a 10 10))
	   (n1ptr (mix-number-start num1 10)) 
	   (len1 (- 11 n1ptr))
	   (num2 (mix-copy-number b 5 10))
	   (n2ptr (mix-number-start num2 10))
	   (len2 (- 11 n2ptr))
	   (qval (mix-make-number 10))
	   (qdigits (cond
		     ((> len2 len1)
		      1)
		     (t
		      (1+ (- len1 len2))))))
      (dec n1ptr) ;; FIXME: possible overflow
      (cond
       ((<= len2 len1)
	;;; Actual divide algorythm starts here
	;;; See Knuth TAOCP Vol 2. p237.
	;;;---------------------------------------
	;; Normalize
	(let ((norm (/ +BASE+ (1+ (mix-byte num2 n2ptr)))))
	  (if (> norm 1)
	      (setq num1 (mix-one-mul num1 norm 10)
		    num2 (mix-one-mul num2 norm 10))))
	;; Initialize divide loop
	(let ((qdig n1ptr)          ;; 1+number of quot.digits collected so far
	      (qptr (- 11 qdigits)));; position of current byte in quotient
	  (while (< (- qdig n1ptr) qdigits) ;; FIXME1
	    (let ((borrow 0)
		  ;; Calculate the quotient digit guess
		  (qguess (cond
			   ((= (mix-byte num2 n2ptr) (mix-byte num1 qdig))
			    9)
			   (t
			    (/ (+
				(* (mix-byte num1 qdig) +BASE+)
				(mix-byte num1 (1+ qdig)))
			       (mix-byte num2 n2ptr))))))
	      ;; test qguess
	      (cond
	       ((test-guess num1 num2 qdig qguess n2ptr)
		(dec qguess)
		;; repeat
		(cond
		 ((test-guess num1 num2 qdig qguess n2ptr)
		  (dec qguess)))))
	      ;; Multiply and subtract
	      (cond
	       ((> qguess 0)
		(let* ((mval (mix-one-mul num2 qguess 10))
		       (ptr1 (+ qdig len2))
		       (ptr2 10)
		       (count 0))
		  (while (< count (1+ len2))
		    (let ((val (- (mix-byte num1 ptr1)
				  (mix-byte mval ptr2)
				  borrow)))
		      (cond
		       ((< val 0)
			(setq val (+ val +BASE+))
			(setq borrow 1))
		       (t
			(setq borrow 0)))
		      (set-byte num1 ptr1 val)
		      (dec ptr1)
		      (dec ptr2)
		      (inc count))))))
	      ;; Test for negative result
	      (cond
	       ((= borrow 1)
		(dec qguess)
		(let ((ptr1 (+ qdig len2))
		      (ptr2 (1- (+ n2ptr len2)))
		      (carry 0)
		      (count 0))
		  (while (< count len2)
		    (let ((val (+ (mix-byte num1 ptr1)
				  (mix-byte num2 ptr2)
				  carry)))
		      (cond
		       ((> val 9)
			(setq val (- val +BASE+))
			(setq carry 1))
		       (t
			(setq carry 0)))
		      (set-byte num1 ptr1 val))
		    (inc count))
		  (if (= carry 1)
		      (set-byte num1 ptr1
				(% (1+ (mix-byte num1 ptr1))
				   +BASE+))))))
	      (set-byte qval qptr qguess))
	    (inc qptr)
	    (inc qdig)))))
      (mix-copy-number qval 10 5)))))

;; Returns ( QUOT . REM )
(defun mix-unsigned-div-mod (a b)
  (let ((quot (mix-unsigned-div a b)))
    (cons
     quot
     (mix-copy-number 
      (mix-unsigned-sub a
			(mix-unsigned-mul quot b) 10) 10 5))))

    
;;; Signed arithmetics
;; Comparison
(defun mix-eq (a b)
  (and (= (mix-sign a) (mix-sign b))
       (mix-unsigned-eq a b)))

(defun mix-ne (a b)
  (not (mix-eq a b)))

(defun mix-gt (a b)
  (let ((sign-a (mix-sign a))
	(sign-b (mix-sign b)))
    (cond
     ((eq sign-a sign-b)
      (if (eq sign-a 1)
	  (mix-unsigned-gt a b)
	(mix-unsigned-gt b a)))
     ((eq sign-a 1)
      t)
     (t
      nil))))

(defun mix-ge (a b)
  (let ((sign-a (mix-sign a))
	(sign-b (mix-sign b)))
    (cond
     ((eq sign-a sign-b)
      (if (eq sign-a 1)
	  (mix-unsigned-ge a b)
	(mix-unsigned-ge b a)))
     ((eq sign-a 1)
      t)
     (t
      nil))))

(defun mix-lt (a b)
  (not (mix-ge a b)))

(defun mix-le (a b)
  (not (mix-gt a b)))

;; Addition / Subtraction
(defun mix-add (a b)
  (let ((sign-a (mix-sign a))
	(sign-b (mix-sign b)))
    (cond
     ((eq sign-a sign-b)
      (let ((c (mix-unsigned-add a b)))
	(set-sign c sign-a)
	c))
     (t ;; signs are different
      (if (mix-unsigned-lt a b)
	  (let (tmp)
	    (setq tmp b
		  b a
		  a tmp)))
      ;; Now mod(a) >= mod(b)
      (let ((c (mix-unsigned-sub a b)))
	(set-sign c (mix-sign a))
	c)))))

(defun mix-sub (a b)
  (let ((sign-a (mix-sign a))
	(sign-b (mix-sign b))
	(sign 1))
    (cond
     ((eq sign-a sign-b)
      (cond
       ((mix-unsigned-lt a b)
	(let (tmp)
	  (setq tmp b
		b a
		a tmp))
	(setq sign (- (mix-sign a))))
       (t
	(setq sign (mix-sign a))))
      (let ((c (mix-unsigned-sub a b)))
	(set-sign c sign)
	c))
     (t ;; signs are different
      (let ((c (mix-unsigned-add a b)))
	(set-sign c sign-a)
	c)))))

;; Multiplication / Division
(defun mix-mul (a b)
  (let ((c (mix-unsigned-mul a b)))
    (set-sign c (* (mix-sign a) (mix-sign b)))
    c))

(defun mix-div-mod (a b)		
  (let ((result (mix-unsigned-div-mod a b)))
    (set-sign (cdr result) (mix-sign a))
    (set-sign (car result) (* (mix-sign a) (mix-sign b)))
    result))

;; Memory Dump and abort routines

(defun mix-sign-char (value)
  (cond
   ((> (mix-sign value) 0) 
    ?+)
   (t
    ?-)))

;; Format the MIX-word as a number
(defun mix-format-number (value)
  (let ((num (mix-copy-number value 5 10))
	(string (make-string 14 ?0))
	(i 13))
    (while (and (> i 0)
		(not (mix-zero-p num 10)))
      (let ((r (mix-unsigned-div-mod num '[ 1 0 0 0 0 10 ])))
	(aset string i (string-to-char
			(number-to-string (mix-byte (cdr r) 5))))
	(setq num (mix-copy-number (car r) 5 10)))
      (dec i))
    (aset string 0 (mix-sign-char value))
    string))

;; Format MIX word as a cell in TAOCP
;;ADD I F C
;;1 2 3 4 5
(defun mix-format-cell (value)
  (let ((string (make-string 14 ?0))
	(addr (mix-get-addr value))
	(i    (mix-index value))
	(f    (mix-get-fld value))
	(c    (mix-get-code value)))
    (format "%c%04d %02d %02d %02d"
	    (mix-sign-char value)
	    addr i f c)))

;; Format mix cell as a ALF
(defun mix-format-alpha (value)
  (let ((string (make-string 5 ?\ ))
	(i 1))
    (while (<= i 5)
      (aset string (1- i) (mix-xlat-to-ascii (mix-byte value i)))
      (inc i))
    string))

;; Format MIX word as a string of bytes
(defun mix-format-bytes (value)
  (format "%c%02d %02d %02d %02d %02d"
	  (mix-sign-char value)
	  (mix-byte value 1)
	  (mix-byte value 2)
	  (mix-byte value 3)
	  (mix-byte value 4)
	  (mix-byte value 5)))

(defun mix-format-ireg-bytes (num)
  (let ((val (mix-get-reg num)))
    (format "%c%02d %02d"
	    (mix-sign-char val)
	    (mix-byte val 4)
	    (mix-byte val 5))))

(defun mix-format-ireg-number (num)
  (let ((val (mix-get-reg num)))
    (format "%c%05d"
	    (mix-sign-char val)
	    (+ (* (mix-byte val 4) +BASE+)
	       (mix-byte val 5)))))
     
;; Dump state of registers
(defun mix-dump-registers ()
  (let ((areg (mix-get-reg :rA))
	(xreg (mix-get-reg :rX)))
    (insert
     (format "Registers A/X    %s   %s\n"
	     (mix-format-bytes areg)
	     (mix-format-bytes xreg))
     (format "                 %s    %s\n\n"
	     (mix-format-number areg)
	     (mix-format-number xreg))
     (format "Index Registers  %6.6s %6.6s %6.6s %6.6s %6.6s %6.6s\n"
	     (mix-format-ireg-bytes 1)
	     (mix-format-ireg-bytes 2)
	     (mix-format-ireg-bytes 3)
	     (mix-format-ireg-bytes 4)
	     (mix-format-ireg-bytes 5)
	     (mix-format-ireg-bytes 6))
     (format "                 %6.6s %6.6s %6.6s %6.6s %6.6s %6.6s\n"
	     (mix-format-ireg-number 1)
	     (mix-format-ireg-number 2)
	     (mix-format-ireg-number 3)
	     (mix-format-ireg-number 4)
	     (mix-format-ireg-number 5)
	     (mix-format-ireg-number 6))
     "\n"

     (format
      "Jump Register    %6d     Overflow toggle:      %s\n"
      (mix-convert-word-to-number (mix-get-reg :rJ))
      (cond
       (*mix-ovtog* "ON")
       (t           "OFF")))
     
     (format
      "                 %6d     Comparison Indicator: %s\n"
      (mix-convert-word-to-number (mix-get-reg :rJ))
      (cond
       ((= *mix-compi* 0)  "EQUAL")
       ((< *mix-compi* 0)  "LESS")
       (t                  "GREATER")))

     "\n"

     (format
      "Clock = %6d u. Loc = %4d, M %2d, I %2d, F %2d, C %2d, inst = %s\n"
      *mix-uclock*
      *mix-loc*
      *mix-mem*
      (mix-index *mix-inst*)
      *mix-fld*
      *mix-op*
      (mix-format-number *mix-inst*))

     "\n" )))

(defun mix-format-memory (fun loc count)
  (let ((i 0))
    (while (< i count)
      (funcall fun (mix-cell (+ loc i)))
      (inc i))))

(defconst +WPL+ 4)

(defun mix-mem-eq (loc1 loc2 count)
  (let ((i 0)
	(rc nil))
    (while (and (< i count)
		(setq rc (mix-eq (mix-cell (+ loc1 i))
				 (mix-cell (+ loc2 i)))))
      (inc i))
    rc))

(defun mix-dump-memory ()
  (let ((loc 0)
	(last 0)
	(repeat nil))
    (while (< loc +MEMSIZE+)
      (cond
       ((or (= loc last)
	    (not (mix-mem-eq last loc +WPL+)))
	(cond
	 (repeat
	  (insert
	   (format "    Lines %04d to %04d are the same.\n"
		   last (- loc +WPL+)))
	  (setq repeat nil)))

	;; Numbers
	(insert (format "%04d " loc))
	(let ((j 0))
	  (while (< j +WPL+)
	    (insert
	     (format " %s"
		     (mix-format-number (mix-cell (+ loc j)))))
	    (inc j)))
	;; Cells
	(insert "\n     ")
	(let ((j 0))
	  (while (< j +WPL+)
	    (insert
	     (format " %s"
		     (mix-format-cell (mix-cell (+ loc j)))))
	    (inc j)))
	;; ALF
	(insert "\n     ")
	(let ((j 0))
	  (while (< j +WPL+)
	    (insert
	     (format " '%5.5s'       "
		     (mix-format-alpha (mix-cell (+ loc j)))))
	    (inc j)))

	(insert "\n")
	(setq last loc))
       (t
	(setq repeat t)))
      (setq loc (+ loc +WPL+)))
    ;; Final test
    (cond
     (repeat
      (insert
       (format "    Lines %04d to %04d are the same.\n"
	       last (- loc +WPL+)))))))


(defun mix-dump (&optional msg nomem)
  (switch-to-buffer "*mix-state*") ;;FIXME
;;  (delete-region 0 (point-max))
  (if msg
      (insert msg "\n"))
  (mix-dump-registers)
  (if nomem
      ()
    (mix-dump-memory)))


;; Error reporting
(defun mix-memerror ()
  (mix-dump (format "Invalid memory location: %d" *mix-mem*))
  (throw 'mix-abort t))

(defun mix-fielderror ()
  (mix-dump (format "Invalid field: %d" *mix-fld*))
  (throw 'mix-abort t))

(defun mix-addrerror ()
  (mix-dump (format "Invalid address: %d" *mix-mem*))
  (throw 'mix-abort t))

(defun mix-operror ()
  (mix-dump (format "Invalid opcode: %d" *mix-op*))
  (throw 'mix-abort t))

(defun mix-indexerror ()
  (mix-dump (format "Invalid index register: %d" *mix-rI2*))
  (throw 'mix-abort t))

(defun mix-sizeerror ()
  (mix-dump (format "Invalid size: %s" *mix-rA*))
  (throw 'mix-abort t))

;; Stream I/O operations

(defun open-stream-unit ()
  (save-excursion
    (set-buffer (mix-get-unit-buffer))
    (beginning-of-buffer))
  t)

(defun close-stream-unit ()
  )

;; Read data from a stream device (card reader (16) or tty (17)
(defun stream-reader (bufsize)
  (save-excursion
    (set-buffer (mix-get-unit-buffer)) 
    (let* ((cpt (1+ (mix-get-unit-parm :io-position))) ;; Current point
	   (npt (progn                                 ;; Next point
		  (goto-char cpt)
		  (forward-line 1)
		  (point)))
	   (count (- npt cpt 1))                   ;; Number of symbols to read
	   (bufpos 0)                              ;; Position in buffer
	   (address (mix-get-unit-parm :io-address))) ;; I/O memory address
      (cond
       ((> npt cpt)
	;; Set new i/o position
	(mix-set-unit-parm :io-position (1- npt))
	;; Read and convert memory contents
	(while (< bufpos count) ;; WAS: bufsize)
	  (let ((cellval (vector 1 0 0 0 0 0))
		(j 0))
	    ;; Read next +BYTESPERWORD+ bytes and pack them into a word
	    (while (< j +BYTESPERWORD+)
	      (let* ((p (+ (* bufpos +BYTESPERWORD+) j))
		     (c (and
			 (< p count) ;; WAS: bufsize)
			 (mix-xlat-to-mix
			  (string-to-char
			   (buffer-substring (+ cpt p) (+ cpt p 1)))))))
		(if c
		    (aset cellval (1+ j) c)))

	      (inc j))
		
	    (mix-set-cell address cellval)   ;; Store into memory
	    (inc address))                   ;; next memory address
	  (inc bufpos))                      ;; next buffer position
	(mix-set-unit-parm :io-address address)
	t)
       (t
	nil)))))

(defun stream-writer (bufsize)
  (save-excursion
    (set-buffer (mix-get-unit-buffer)) 
    (let ((string (make-string (* bufsize +BYTESPERWORD+) ?\ ))
	  (address (mix-get-unit-parm :io-address))
	  (pos 0)
	  (word 0))
      (while (< word bufsize)
	(let ((byte 0)
	      (cell (mix-get-cell (+ address word))))
	  (while (< byte +BYTESPERWORD+)
	    (aset string pos
		  (mix-xlat-to-ascii (mix-byte cell (1+ byte))))
	    (inc pos)
	    (inc byte)))
	(inc word))
      (insert string "\n"))))
  

(defun mix-reader ()
  (let ((io-op (mix-get-unit-parm :io-op)))
    (cond
     ((eq io-op +IOC+)
      )
     ((eq io-op +IN+)
      (and (open-stream-unit)
	   (stream-reader 80)))
     ((eq io-op +OUT+)
      ))))

(defun mix-printer ()
  (open-stream-unit)
  (let ((io-op (mix-get-unit-parm :io-op)))
    (cond
     ((eq io-op +IOC+)
      (mix-set-unit-parm :io-clock (+ *mix-uclock*
				      (mix-get-unit-parm :io-seek-time)))
      (insert "\f"))
     ((eq io-op +IN+)
      (message "Invalid operation for unit 18 (PRINTER)"))
     ((eq io-op +OUT+)
      (stream-writer 24)))))

(defun mix-tape ()
  (let ((io-op (mix-get-unit-parm :io-op)))
    (cond
     ((eq io-op +IOC+)
      (mix-set-unit-parm :io-clock
			 (+ *mix-uclock*
			    (*
			     (mix-get-unit-parm :io-seek-time)
			     (cond
			      ((= *mix-mem* 0)
			       (mix-get-unit-parm :io-position))
			      (t
			       (abs *mix-mem*)))))))
     ((eq io-op +IN+)
      )
     ((eq io-op +OUT+)
      ))))

(defun mix-disk ()
  (let ((io-op (mix-get-unit-parm :io-op)))
    (cond
     ((eq io-op +IOC+)
      (mix-set-unit-parm :io-clock
			 (+ *mix-uclock*
			    (*
			     (abs (- (mix-get-reg :rX)
				     (mix-get-unit-parm :io-position)))
			     (mix-get-unit-parm :io-seek-time))))
      (mix-set-unit-parm :io-seek-pos (mix-get-reg :rX)))
     ((eq io-op +IN+)
      )
     ((eq io-op +OUT+)
      ))))

(defun mix-do-io ()
  (cond
   ((not (mix-get-unit-parm :io-complete))
    (funcall (mix-get-unit-parm :io-ufn))
    (mix-set-unit-parm :io-complete t)
    (mix-set-unit-parm :io-clock (+ *mix-uclock*
				    (/
				     (mix-get-unit-parm :io-time)
				     2))))
   (t
    (mix-set-unit-parm :io-clock -1))))

(defun mix-do-outstanding-io (clock)
  (let ((fld *mix-fld*)
	(next 0))
    (while (not (= next *mix-instime*))
      (setq next *mix-instime*)
      (let ((i 0))
	(while (< i +NUMUNITS+)
	  (let ((clock (aref (aref *mix-unit* i) :io-clock)))
	    (cond
	     ((and (> clock 0) (< clock next))
	      (setq next clock)
	      (setq fld i))))
	  (inc i))
	(cond
	 ((not (= next *mix-instime*))
	  (mix-select-unit fld)
	  (setq *mix-uclock* (mix-get-unit-parm :io-clock))
	  (mix-do-io))))))
  (mix-select-unit *mix-fld*))
    
    
;; ---------
;; Basic MIX instructions

(defconst +LDAN+ 16)
(defconst +LDA+  8)
(defconst +STA+  24)
(defconst +INCA+ 48)

;; Memory fetch routine
;; Returns contents of memory location +MIX-MEM+
(defun mix-memory ()
  (if (or (< *mix-mem* 0) (>= *mix-mem* +MEMSIZE+))
      (mix-memerror))
  (let ((src (mix-cell *mix-mem*))
	(i 0))
    (while (<= i 5)
      (set-byte *mix-rA* i (mix-byte src i))
      (inc i))))
    
;; The fcheck subroutine process a partial field specification,
;; making sure that it has the form 8*L+R with L <= R <= 5.
;; Returns value of L.
(defun mix-fcheck ()
  (let ((q (/ *mix-fld* 8))
	(r (% *mix-fld* 8)))
    (setq *mix-rI1* q
	  *mix-rA*  (mix-convert-to-word q)
	  *mix-rI2* r
	  *mix-rX*  (mix-convert-to-word r))
    (if (or (> q r)
	    (> r 5))
	(mix-fielderror))))

(defun mix-sizechk ()
  (cond
   (*mix-iflag*
    (if (mix-gt (mix-abs *mix-rA*) '[1 0 0 0 63 63])
	(mix-sizeerror)))
   (t
    (mix-ovcheck))))

(defun mix-ovcheck ()
  ;; FIXME: not needed?
  )

;; Normally called by GETV to extract a field from memory.
;; GETAV is called directly only in comparison operations
;; to extract a field from a register.
;; Input: rI1 = L
;;        rI2 = R
;; Output:rA = V
;;        rI1, rI2 preserved
(defun mix-getav ()
  (if (= *mix-rI1* 0)    
      (inc *mix-rI1*)          ;; Preserve the sign if it is included in the
    (set-sign *mix-rA* 1)) ;; field, otherwise make it positive
  (let ((src-ind *mix-rI2*)
	(dst-ind 5))
    (while (>= src-ind *mix-rI1*)
      (set-byte *mix-rA* dst-ind (mix-byte *mix-rA* src-ind))
      (dec src-ind)
      (dec dst-ind))
    (while (> dst-ind 0)
      (set-byte *mix-rA* dst-ind 0)
      (dec dst-ind))))
  
;; getv finds the quantity V (namely, the appropriate field
;; of location M) used in various MIX operators.
(defun mix-getv ()
  (mix-fcheck) ;; Process the field and set rI1 <- L
  (mix-memory) ;; rA <- signed memory value 
  (mix-getav)) ;; Extract the field

(defun mixop-nop ()
  (if (= *mix-fld* 63) ;; Use NOP to reset the clock if F is 63.
      (setq *mix-uclock* -1)))

(defun mixop-add ()
  (mix-getv)           ;; Get the value V in rA and rX.
  (setq *mix-rI1* 0)   ;; Let rI1 indicate the A register.
  (mix-inc))

(defun mixop-sub ()
  (mix-getv)
  (setq *mix-rI1* 0)
  (mix-dec))

(defun mixop-mul ()
  (mix-getv)
  (let ((prod (mix-mul *mix-rA* (mix-get-reg :rA))))
    (mix-set-reg :rX (mix-copy-number (mix-rsh prod 5 10) 10 5))
    (mix-set-reg :rA (mix-copy-number prod 10 5))))

(defun mixop-div ()
  (mix-getv)
  (let ((res (mix-div-mod (mix-unsigned-add
			   (mix-lsh (mix-copy-number (mix-get-reg :rX) 5 10)
				    5 10)
			   (mix-copy-number (mix-get-reg :rA) 5 10))
			  *mix-rA*)))
    (mix-set-reg :rA (car res))
    (mix-set-reg :rX (cdr res))))


(defun mixop-loadn ()
  (mix-getv)
  (setq *mix-rI1* (- *mix-op* +LDAN+))
  (mix-setq (mix-get-reg *mix-rI1*) 5 *mix-rA* 5)
  (mix-sizechk))

(defun mixop-load ()
  (mix-getv)
  (setq *mix-rI1* (- *mix-op* +LDA+))
  (mix-setq (mix-get-reg *mix-rI1*) 5 *mix-rA* 5)
  (mix-sizechk))

(defun mixop-store ()
  (mix-fcheck)       ;; rI1 <- L
  (mix-memory)       ;; Get the contents of the memory location.
  (let ((regval (mix-get-reg (- *mix-op* +STA+)))  ;; register value
	(dst-ind *mix-rI2*)                       ;; destination index
	(src-ind 5))                              ;; source index
    (if (= *mix-rI1* 0)
	(progn
	 (set-sign (mix-cell *mix-mem*) (mix-sign regval))
	 (inc *mix-rI1*)))
    (while (>= dst-ind *mix-rI1*)
      (set-byte (mix-cell *mix-mem*) dst-ind (mix-byte regval src-ind))
      (dec src-ind)
      (dec dst-ind))))

(defconst *mix-jumptab*
  '[;; 0 -- jmp
    (lambda (regval)
      (mix-jmp))
    ;; 1 -- jsj
    (lambda (regval)
      (mix-jsj))
    ;; 2 -- jov
    (lambda (regval)
      (let ((ov *mix-ovtog*))
	(setq *mix-ovtog* nil)
	(if ov
	    (mix-jmp))))
    ;; 3 -- jnov
    (lambda (regval)
      (let ((ov *mix-ovtog*))
	(setq *mix-ovtog* nil)
	(if (not ov)
	    (mix-jmp))))
    ;; 4 -- jlt
    (lambda (regval)
      (if (mix-negative-p regval)
	  (mix-jmp)))
    ;; 5 -- jeq
    (lambda (regval)
      (if (mix-zero-p regval)
	  (mix-jmp)))
    ;; 6 -- jgt
    (lambda (regval)
      (if (mix-positive-p regval)
	  (mix-jmp)))
    ;; 7 -- jge
    (lambda (regval)
      (if (mix-nonnegative-p regval)
	  (mix-jmp)))
    ;; 8 -- jne
    (lambda (regval)
      (if (not (mix-zero-p regval))
	  (mix-jmp)))
    ;; 9 -- jle
    (lambda (regval)
      (if (mix-nonpositive-p regval 0)
	  (mix-jmp))) ])

(defun mixop-jump ()
  (if (> *mix-fld* 9)
      (mix-fielderror))
  (funcall (aref *mix-jumptab* *mix-fld*) *mix-rA*))

(defun mix-jmp ()
  (mix-set-reg :rJ (mix-convert-to-word *mix-loc*))
  (mix-jsj))
       
(defun mix-jsj ()
  (mix-memory)
  (setq *mix-loc* *mix-mem*))

(defun mixop-regjump ()
  (let ((regval (mix-get-reg (- *mix-op* +JA+))))
    (if (> *mix-fld* 5)
	(mix-fielderror))
    (funcall (aref *mix-jumptab* (+ *mix-fld* 4)) regval)))

(defun mixop-addrop ()
  (if (> *mix-fld* 3)
      (mix-fielderror))
  (setq *mix-rA* (mix-convert-to-word *mix-mem*))
  ;; When M is zero ENTA loads the sign of
  ;; the instruction and ENNA loads the opposite sign.
  (if (= *mix-mem* 0)
      (mix-copy-sign *mix-rA* *mix-inst*))
  (setq *mix-rI1 (- *mix-op* +INCA+))
  (funcall (aref '[;; 0
		   mix-inc
		   ;; 1
		   mix-dec
		   ;; 2 -- enta
		   (lambda ()
		     (mix-set-reg *mix-rI1* (copy-sequence *mix-rA*))
		     (mix-sizechk))
		   ;; 3 -- enna
		   (lambda ()
		     (let ((tmp (copy-sequence *mix-rA*)))
		       (set-sign tmp (- mix-sign tmp))
		       (mix-set-reg *mix-rI1* tmp)
		       (mix-sizechk))) ]
		 *mix-fld*)))

(defun mix-dec ()
  (mix-set-reg *mix-rI1* (mix-sub
			  (mix-get-reg *mix-rI1*)
			  *mix-rA*))
  (mix-sizechk))

(defun mix-inc ()
  (mix-set-reg *mix-rI1* (mix-add
			  (mix-get-reg *mix-rI1*)
			  *mix-rA*))
  (mix-sizechk))

(defun mixop-compare ()
  (let (a)
    (getv)
    (setq a (copy-sequence *mix-rA*))
    (getav)
    (setq *mix-compi*
	  (cond
	   ((mix-eq *mix-rA* a)
	    0)
	   ((mix-gt *mix-rA* a)
	    1)
	   (t
	    -1)))))
      
(defun mixop-shift ()
  (if (< *mix-mem* 0)
      (mix-addrerror))
  (if (> *mix-fld* 5)
      (mix-fielderror))
  (funcall (aref '[;; 0 -- sla
		   (lambda ()
		     (mix-set-reg :rA
				  (mix-lsh (mix-get-reg :rA) *mix-mem*)))
		   ;; 1 -- sra
		   (lambda ()
		     (mix-set-reg :rA
				  (mix-rsh (mix-get-reg :rA) *mix-mem*)))
		   ;; 2 -- slax
		   (lambda ()
		     (let ((temp (mix-lsh
				  (mix-unsigned-add
				   (mix-lsh
				    (mix-copy-number (mix-get-reg :rA) 5 10)
				    5 10)
				   (mix-copy-number (mix-get-reg :rX) 5 10))
				  *mix-mem*
				  10)))
		       (mix-set-reg-magn :rX (mix-copy-number temp 10 5))
		       (mix-set-reg-magn :rA (mix-copy-number
					      (mix-rsh temp 5 10)
					      10 5))))
		   ;; 3 -- srax
		   (lambda ()
		     (let ((temp (mix-rsh
				  (mix-unsigned-add
				   (mix-lsh
				    (mix-copy-number (mix-get-reg :rA) 5 10)
				    5 10)
				   (mix-copy-number (mix-get-reg :rX) 5 10))
				  *mix-mem*
				  10)))
		       (mix-set-reg-magn :rX (mix-copy-number temp 10 5))
		       (mix-set-reg-magn :rA (mix-copy-number
					      (mix-rsh temp 5 10)
					      10 5))))
		   ;; 4 -- slc
		   (lambda ()
		     (let ((temp (mix-lcsh
				  (mix-unsigned-add
				   (mix-lsh
				    (mix-copy-number (mix-get-reg :rA) 5 10)
				    5 10)
				   (mix-copy-number (mix-get-reg :rX) 5 10))
				  *mix-mem*
				  10)))
		       (mix-set-reg-magn :rX (mix-copy-number temp 10 5))
		       (mix-set-reg-magn :rA (mix-copy-number
					      (mix-rsh temp 5 10)
					      10 5))))
		   ;; 5 -- src
		   (lambda ()
		     (let ((temp (mix-rcsh
				  (mix-unsigned-add
				   (mix-lsh
				    (mix-copy-number (mix-get-reg :rA) 5 10)
				    5 10)
				   (mix-copy-number (mix-get-reg :rX) 5 10))
				  *mix-mem*
				  10)))
		       (mix-set-reg-magn :rX (mix-copy-number temp 10 5))
		       (mix-set-reg-magn :rA (mix-copy-number
					      (mix-rsh temp 5 10)
					      10 5)))) ]
	   *mix-fld*)))

(defun mixop-move ()
  (let ((i1reg (mix-convert-word-to-number (mix-get-reg :rI1))))
    (while (> *mix-fld* 0)
      (mix-memory)
      (if (< i1reg 0)
	  (mix-memerror))
      (if (> i1reg +MEMSIZE+)
	  (mix-memerror))
      (mix-set-cell i1reg (copy-sequence *mix-rA*))
      (setq *mix-uclock* (+ *mix-uclock* 2))
      (inc i1reg)
      (inc *mix-mem*)
      (dec *mix-fld*))
    (mix-set-reg-magn :rI1 (mix-convert-to-word i1reg))))

(defun mixop-jbus ()
  (cond
   ((> (mix-get-unit-parm :io-clock) *mix-uclock*)
    (cond
     ((= (1- *mix-loc*) *mix-mem*)
      (mix-do-outstanding-io (mix-get-unit-parm :io-clock))
      (setq *mix-uclock* (mix-get-unit-parm :io-clock))))
    (mix-jmp))))

(defun mixop-ioc ()
  (cond
   ((or (= *mix-fld* +PUNCH+) (= *mix-fld* +READER+))
    )
   (t
    (cond
     ((> (mix-get-unit-parm :io-clock) *mix-uclock*)
      (mix-do-outstanding-io (mix-get-unit-parm :io-clock))
      (setq *mix-uclock* (mix-get-unit-parm :io-clock))
      (mix-do-io)))
    ;; again...
    (cond
     ((> (mix-get-unit-parm :io-clock) *mix-uclock*)
      (mix-do-outstanding-io (mix-get-unit-parm :io-clock)) 
      (setq *mix-uclock* (mix-get-unit-parm :io-clock))
      (mix-do-io)))
    (mix-set-unit-parm :io-op +IOC+)
    (mix-set-unit-parm :io-address *mix-mem*)
    (funcall (mix-get-unit-parm :io-ufn)))))
	     
(defun mixop-in ()
    (cond
     ((> (mix-get-unit-parm :io-clock) *mix-uclock*)
      (mix-do-outstanding-io (mix-get-unit-parm :io-clock))
      (setq *mix-uclock* (mix-get-unit-parm :io-clock))
      (mix-do-io)))
    (cond
     ((> (mix-get-unit-parm :io-clock) *mix-uclock*)
      (mix-do-outstanding-io (mix-get-unit-parm :io-clock))
      (setq *mix-uclock* (mix-get-unit-parm :io-clock))
      (mix-do-io)))
    (mix-set-unit-parm :io-op +IN+)
    (mix-set-unit-parm :io-address *mix-mem*)
    (mix-set-unit-parm :io-clock (+ *mix-uclock*
				    (/ (mix-get-unit-parm :io-time) 2)))
    (mix-set-unit-parm :io-complete nil)
    (cond
     ((and (>= *mix-fld* +DISK8+) (<= *mix-fld* +DISK15+))
      (mix-set-unit-parm :io-clock (abs
				    (*
				     (- (mix-get-reg :rX)
					(mix-get-unit-parm :io-position))
				     (mix-get-unit-parm :io-seek-time))))
      (mix-set-unit-parm :io-seek-pos (mix-get-reg :rX)))))

;; Convert a literal string from rA:rX into number and store it into rA
;; FIXME: check whether ovtog should be set
(defun mixop-num ()
  (let ((string (mix-unsigned-add (mix-lsh
				   (mix-copy-number (mix-get-reg :rA) 5 10)
				   5 10)
				  (mix-copy-number (mix-get-reg :rX) 5 10)
				  10))
	(temp (mix-make-number 10))
	(i 1))
    (while (<= i 10)
      (let* ((num (% (mix-byte string i) 10)))
	(setq temp (mix-unsigned-add
		    (mix-one-mul temp 10 10)
		    (mix-convert-to-word num 10)
		    10)))
      (inc i))
    (if (mix-unsigned-gt temp '[ 1 0 0 0 0 0 63 63 63 63 63 ])
	(setq *mix-ovtog* t))
    (mix-set-reg-magn :rA (mix-copy-number temp 10 5))))

;; Convert a number from rA into literal and store the result into rA:rX
(defun mixop-char ()
  (let ((num (mix-copy-number (mix-get-reg :rA) 5 10))
	(string (mix-make-number 10))
	(i 10))
    (while (> i 0)
      (let ((r (mix-unsigned-div-mod num '[ 1 0 0 0 0 10 ])))
	(set-byte string i (+ 30 (mix-byte (cdr r) 5)))
	(setq num (mix-copy-number (car r) 5 10)))
      (dec i))
    (mix-set-reg-magn :rA (mix-copy-number (mix-rsh string 5 10) 10 5))
    (mix-set-reg-magn :rX (mix-copy-number string 10 5))))

(defun mixop-hlt ()
  (mix-do-outstanding-io 99999)
  (setq *mix-halt* t))
		   
(defun mixop-spec ()
  (funcall (aref '[;; 0 -- num
		   mixop-num
		   ;; 1 -- char
		   mixop-char
		   ;; 2 -- hlt
		   mixop-hlt]
		 *mix-fld*)))

		   
;; ---------

(defun mix-init-memory ()
  (mix-fill-array *mix-cell* +MEMSIZE+ (vector 1 0 0 0 0 0)))

(defun mix-init-registers ()
  (mix-fill-array *mix-reg* 10 (vector 1 0 0 0 0 0))
  (setq *mix-compi* 0
	*mix-ovtog* nil
	*mix-uclock* 0
	*mix-instime* 0
	*mix-rA* (mix-make-number 5)
	*mix-rX* 0
	*mix-rI1* 0
	*mix-rI2* 0))

(defun mix-init-io ()
  (let ((i 0))
    (while (< i +NUMUNITS+)
      (let* ((unit (aref *mix-unit* i)))
	(aset unit :io-position 0)
	(aset unit :io-tape-end 0)
	(aset unit :io-address 0)
	(aset unit :io-op +IOC+)
	(aset unit :io-complete t)
	(aset unit :io-buffer nil))
      (inc i))))

(defun mix-init()
  (mix-init-memory)
  (mix-init-registers)
  (mix-init-io))

(defun mix-run ()
  (setq *mix-halt* nil)
  (while (not *mix-halt*)
    (setq *mix-uclock* *mix-instime*)
    ;; Check pending i/o
    (let ((fld 0))
      (while (< fld +NUMUNITS+)
	(mix-select-unit fld)
	(mix-do-io)
	(inc fld)))
      ;;;;
    (setq *mix-inst* (mix-cell *mix-loc*))
    (setq *mix-mem*  (mix-get-addr *mix-inst*))
    (setq *mix-op*   (mix-get-code *mix-inst*))
    (setq *mix-fld*  (mix-get-fld *mix-inst*))
    (inc *mix-loc*)
    (if (> *mix-loc* +MEMSIZE+)
	(mix-addrerror))
    (if (< (mix-sign *mix-inst*) 0)
	(setq *mix-mem* (- *mix-mem*)))
    (setq *mix-rI2* (mix-index *mix-inst*))
    (cond
     ((> *mix-rI2* 0)
      (if (> *mix-rI2* 6)
	  (mix-indexerror))
      (setq *mix-mem* (+ *mix-mem*
			 (mix-convert-word-to-number (mix-get-reg *mix-rI2*))))
      (if (> (abs *mix-mem*) 4000)
	  (mix-addrerr))))
    (if (< *mix-fld* +NUMUNITS+)
	(mix-select-unit *mix-fld*))
    (if (> *mix-op* 64)
	(mix-operror))
    (let ((opent (aref +mix-optable+ *mix-op*)))
      (setq *mix-instime* (aref opent :op-time))
      (setq *mix-iflag* (aref opent :op-iflag))
      (funcall (aref opent :op-fun)))))

(defun mix-go (&optional filename)
  (mix-init)
  (mix-select-unit +READER+)
  (if filename
      (mix-set-unit-parm :io-path filename))
  (mix-set-unit-parm :io-op +IOC+)
  (mix-reader)
  (setq uclock (mix-get-unit-parm :io-time))
  (mix-set-unit-parm :io-op +IN+)
  (mix-reader)
  (setq *mix-loc* 0)
  (catch 'mix-abort
    (message "MIX running...")
    (mix-run))
  (message "MIX done")
  (mix-dump nil t)
  t)




